local chunkManager = {
  colors = {
    highlight = {r = 0.5, g = 0.5, b = 0.5, a = 0.1},
    transparent = {0, 0, 0, 0}
  },
  line_width = 1,
  map = {
    "rrrr",
    "grrr",
    "ggrr",
    "ggrg",
    "gggr",
    "gggg",
    "grgr",
    "grgg",
    "grrg",
    "rgrr",
    "rgrg",
    "rggr",
    "rggg",
    "rrgr",
    "rrgg",
    "rrgr",
    "rrrg"
  }
}

local shared = require("shared")
local highlight_chunks = {}

local function get_key_for_value(t, value)
  for k, v in pairs(t) do
    if v == value then
      return k
    end
  end
  return nil
end

function chunkManager.setup_events()
  Event.register(defines.events.on_chunk_generated, chunkManager.on_generated)
  Event.register(defines.events.on_player_mined_entity, chunkManager.on_player_mined_entity)
  Event.register(defines.events.on_marked_for_deconstruction, chunkManager.on_marked_for_deconstruction)
  Event.register({defines.events.on_built_entity, defines.events.on_robot_build_entity}, chunkManager.on_build_entity)
end

function chunkManager.draw(event, register)
  if (chunkManager.isBought(event.position)) then
    return
  end

  local surface = event.surface

  local entities =
    surface.find_entities_filtered {
    area = {{event.position.x * 32, event.position.y * 32}, {event.position.x * 32 + 32, event.position.y * 32 + 32}},
    name = {shared.entity.chunk_entity}
  }

  if entities ~= nil then
    for _, ent in pairs(entities) do
      ent.destroy()
    end
  end

  local b =
    surface.create_entity {
    name = shared.entity.chunk_border,
    position = {event.position.x * 32 + 16, event.position.y * 32 + 16},
    force = "player"
  }

  local str = chunkManager.getChunkMapString(event.position)
  b.graphics_variation = get_key_for_value(chunkManager.map, str.str) or 1

  local c =
    surface.create_entity {
    name = shared.entity.chunk_entity,
    position = {event.position.x * 32 + 16, event.position.y * 32 + 16},
    force = "player"
  }
  c.destructible = false
  c.minable = true
end

function chunkManager.isBought(chunk)
  local bought_chunks = global.control_data.donations.bought
  return bought_chunks[chunk.x] and bought_chunks[chunk.x][chunk.y] == true or false
end

function chunkManager.on_marked_for_deconstruction(event)
  if (event.entity.name == shared.entity.chunk_entity) then
    local player = game.players[event.player_index]
    event.entity.cancel_deconstruction(player.force)

    if
      table.find(
        global.control_data.donations.current_target,
        function(v, k, event)
          return v == event.entity
        end,
        event
      ) == nil
     then
      table.insert(global.control_data.donations.current_target, event.entity)
      chunkManager.highlight(event.entity)
      gui.open_buy_dialog(player)
    end
  end
end

function chunkManager.on_player_mined_entity(event)
  local entity = event.entity
  if entity.name == shared.chunk_entity then
    local chunk = Position.to_chunk_position(entity.position)
    if
      not chunkManager.isBought(chunk) and
        nil ==
          table.find(
            global.control_data.donations.current_target,
            function(v, k, event)
              return v.valid and v.position == event.entity.position
            end,
            event
          )
     then
      local res = {n = entity.name, pos = table.deepcopy(entity.position), f = event.entity.force.name}
      local surface = entity.surface
      entity.destroy()
      local new_ent =
        surface.create_entity {
        name = res.n,
        position = res.pos,
        force = res.f
      }
      new_ent.destructible = false
      new_ent.minable = true
      table.insert(global.control_data.donations.current_target, new_ent)
      gui.open_buy_dialog(game.players[event.player_index], 1)
    end
  end
end

function chunkManager.can_not_build_entity(entity)
  local chunk_position = Position.to_chunk_position(entity.position)
  if chunkManager.isBought(chunk_position) then
    return false
  end

  if entity.type == "car" then
    return false
  end

  if entity.type == "electric-pole" or (entity.type == "entity-ghost" and entity.ghost_type == "electric-pole") then
    local check = entity.position
    local notbuild = true
    for _, x in pairs({-1, 0}) do
      for _, y in pairs({-1, 0}) do
        local new_pos = Position.to_chunk_position({x = check.x + x, y = check.y + y})
        if chunkManager.isBought(new_pos) then
          notbuild = false
        end
      end
    end
    return notbuild
  end

  return true
end

function chunkManager.on_build_entity(event)
  local enabled = settings.global[shared.settings.disable_building].value
  if event.tick < 120 then
    return
  end
  if enabled then
    local entity = event.created_entity
    if chunkManager.can_not_build_entity(entity) then
      local ret = event.robot or game.players[event.player_index]
      local inventory
      if entity.type ~= "entity-ghost" and entity.type ~= "tile-ghost" then
        if event.robot then
          inventory = ret.get_inventory(defines.inventory.robot_cargo)
        else
          inventory = ret.get_inventory(defines.inventory.character_main)
        end
        local name = entity.name
        if name == "straight-rail" or name == "curved-rail" then
          name = "rail"
        end

        if game.item_prototypes[name] and inventory and inventory.can_insert({name = name}) then
          inventory.insert({name = name})
        end
      end

      local inventories_to_check = {"fuel"}
      for _, inv in pairs(inventories_to_check) do
        local inv_s = entity.get_inventory(defines.inventory[inv])
        if inv_s then
          for item, count in pairs(inv_s.get_contents()) do
            inventory.insert {name = item, count = count}
          end
        end
      end
      entity.destroy()
    end
  end
end

function chunkManager.recheckChunks()
  local surface = game.surfaces.nauvis
  chunks = surface.get_chunks()
  for chunk in chunks do
    if surface.is_chunk_generated({x = chunk.x, y = chunk.y}) then
      if chunkRegister[chunk.x] == nil then
        chunkRegister[chunk.x] = {}
        setmetatable(chunkRegister[chunk.x], meta_stringkeys)
      end
      if (chunkRegister[chunk.x][chunk.y] == nil) then
        chunkRegister[chunk.x][chunk.y] = {
          rendered = false
        }
      end

      if (chunkRegister[chunk.x][chunk.y].rendered == false) then
        chunkManager.draw(
          {area = Chunk.to_area(chunk), surface = surface, position = chunk},
          chunkRegister[chunk.x][chunk.y]
        )
        chunkRegister[chunk.x][chunk.y].rendered = true
      end
    end
  end
end

function chunkManager.get_price(chunks)
  local targets = global.control_data.donations.current_target
  local setting = settings.global[shared.settings.chunk_cost].value

  for _, ent in pairs(targets) do
  end
end

function chunkManager.buy(chunk, surface)
  surface = surface or game.surfaces.nauvis
  if chunkRegister[chunk.x] == nil or chunkRegister[chunk.y] == nil then
    -- sanity check, redo all chunks
    chunkManager.recheckChunks()
  end

  local bought_chunks = global.control_data.donations.bought

  if (bought_chunks[chunk.x] == nil) then
    bought_chunks[chunk.x] = {}
    setmetatable(bought_chunks[chunk.x], meta_stringkeys)
  end
  if (bought_chunks[chunk.x][chunk.y] == nil) then
    bought_chunks[chunk.x][chunk.y] = true
  end

  -- remove entities
  local top = chunkManager.getChunkMapString({x = chunk.x - 1, y = chunk.y})
  local left = chunkManager.getChunkMapString({x = chunk.x, y = chunk.y - 1})
  local right = chunkManager.getChunkMapString({x = chunk.x + 1, y = chunk.y})
  local bottom = chunkManager.getChunkMapString({x = chunk.x, y = chunk.y + 1})

  for _, c in pairs({top, left, right, bottom}) do
    local variant = get_key_for_value(chunkManager.map, c.str)
    local pos = Position.from_chunk_position(c.chunk)
    pos.x = pos.x + 16
    pos.y = pos.y + 16
    local borders = surface.find_entities_filtered({position = pos, name = shared.chunk_border})

    if borders then
      for _, border in pairs(borders) do
        border.graphics_variation = variant
      end
    end
  end
end

function chunkManager.getChunkMapString(chunk)
  local str = ""
  if chunkManager.isBought({x = chunk.x, y = chunk.y - 1}) then
    str = str .. "g"
  else
    str = str .. "r"
  end

  if chunkManager.isBought({x = chunk.x + 1, y = chunk.y}) then
    str = str .. "g"
  else
    str = str .. "r"
  end

  if chunkManager.isBought({x = chunk.x - 1, y = chunk.y}) then
    str = str .. "g"
  else
    str = str .. "r"
  end

  if chunkManager.isBought({x = chunk.x, y = chunk.y + 1}) then
    str = str .. "g"
  else
    str = str .. "r"
  end

  return {chunk = chunk, str = str}
end

function chunkManager.process_buy()
  local setting = settings.global[shared.settings.chunk_cost].value
  local donations = global.control_data.donations
  for _, target in pairs(donations.current_target) do
    if target.valid then
      local chunk = Position.to_chunk_position(target.position)
      chunkManager.buy(chunk, target.surface)
      local borders = target.surface.find_entities_filtered({position = target.position, name = shared.chunk_border})
      if borders then
        for _, en in pairs(borders) do
          if en.valid then
            en.destroy()
          end
        end
      end
      target.destroy()
      donations.amount = donations.amount - setting
    end
  end
end

function chunkManager.highlight(entity)
  local pos = Position.from_chunk_position(Position.to_chunk_position(entity.position))
  local location = pos.x .. "," .. pos.y
  if (highlight_chunks[location] == nil) then
    highlight_chunks[location] =
      rendering.draw_rectangle {
      color = chunkManager.colors.highlight,
      filled = true,
      left_top = pos,
      right_bottom = {x = pos.x + 32, y = pos.y + 32},
      draw_on_ground = true,
      surface = entity.surface
    }
  end
end

function chunkManager.unhighlight_all()
  for _, id in pairs(highlight_chunks) do
    if rendering.is_valid(id) then
      rendering.destroy(id)
    end
  end

  highlight_chunks = {}
end

function chunkManager.on_generated(event)
  if (chunkRegister[event.position.x] == nil) then
    chunkRegister[event.position.x] = {}
    setmetatable(chunkRegister[event.position.x], meta_stringkeys)
  end
  if (chunkRegister[event.position.x][event.position.y] == nil) then
    chunkRegister[event.position.x][event.position.y] = {
      rendered = false
    }
  end
  if (chunkRegister[event.position.x][event.position.y].rendered == false) then
    chunkManager.draw(event, chunkRegister[event.position.x][event.position.y])
    chunkRegister[event.position.x][event.position.y].rendered = true
  end
end

return chunkManager
