local mod_gui = require("mod-gui")
local gui = {}
local shared = require("shared")

function math.round(x)
  return x >= 0 and math.floor(x + 0.5) or math.ceil(x - 0.5)
end

function format_num(number, decimal)
  decimal = decimal or 2
  local steps = {
    {1, ""},
    {1e3, "K"},
    {1e6, "M"},
    {1e9, "G"},
    {1e12, "T"}
  }
  for _, b in ipairs(steps) do
    if b[1] <= number + 1 then
      steps.use = _
    end
  end
  local format = "%." .. tostring(decimal) .. "f"
  local result = string.format(format, number / steps[steps.use][1])
  if tonumber(result) >= 1e3 and steps.use < #steps then
    steps.use = steps.use + 1
    result = string.format(format, tonumber(result) / 1e3)
  end
  if steps.use > 1e3 then
    result = string.sub(result, 0, string.sub(result, -1) == "0" and -3 or -1) -- Remove .0 (just if it is zero!)
  end
  return result .. steps[steps.use][2]
end

gui.init_player = function(player, reset)
  gui.get_table(player, reset)
end

gui.collect_totals = function()
  local members = global.control_data.members
  local setting = settings.global[shared.settings.member_spawn_ore].value

  return {
    members = members % setting,
    member_limit = setting
  }
end

gui.update_tables = function()
  totals = gui.collect_totals()
  for _, player in pairs(game.players) do
    gui.update_player_table(player, totals)
  end
end

gui.update_player_table = function(player, totals)
  if player and player.valid and player.gui then
    local guitable = gui.get_table(player)
    if (guitable and guitable.valid) then
      if totals == nil then
        totals = gui.collect_totals()
      end
      guitable[shared.gui.member .. "_value"].caption = totals.members .. " / " .. totals.member_limit

      if guitable[shared.gui.amount_donation .. "_value"] then
        guitable[shared.gui.amount_donation .. "_value"].caption = math.floor(global.control_data.donations.amount)
      end
    end
  end
end

gui.get_table = function(player, reset)
  reset = reset or false
  if not player.gui then
    return
  end
  local frame = player.gui.left
  local info_frame

  if frame == nil or not frame.valid then
    return
  end

  local name_prefix = shared.prefix .. "info_table"
  local settings = settings.get_player_settings(player)

  local setting = settings[shared.settings.show_table].value
  if not setting then
    local flow = frame[name_prefix .. "_flow"]
    if flow and flow.valid then
      flow.destroy()
    end

    return
  end

  local flow = frame[name_prefix .. "_flow"]
  if flow == nil or not flow.valid or reset then
    if flow then
      flow.destroy()
    end
    flow =
      frame.add {
      type = "flow",
      name = name_prefix .. "_flow",
      direction = "vertical"
    }
    flow.style.horizontally_stretchable = false
  end

  info_frame = flow[name_prefix .. "_frame"]
  if info_frame == nil or not info_frame.valid then
    if info_frame then
      info_frame.destroy()
    end
    info_frame =
      flow.add {
      type = "frame",
      name = name_prefix .. "_frame",
      direction = "vertical",
      style = mod_gui.frame_style,
      caption = settings[shared.settings.table_caption].value
    }
    info_frame.style.horizontally_stretchable = false
  end

  local table = info_frame[name_prefix]
  if table == nil then
    table =
      info_frame.add {
      type = "table",
      column_count = 2,
      name = name_prefix
    }
    table.style.column_alignments[1] = "right"

    table.add {
      type = "label",
      name = shared.gui.member .. "_label",
      caption = {"", {shared.gui.member}, " :"},
      style = "bold_label"
    }
    table.add {type = "label", name = shared.gui.member .. "_value", caption = "0 / 0"}

    table.add {
      type = "label",
      name = shared.gui.amount_donation .. "_label",
      caption = {"", {shared.gui.amount_donation}, " (+1 per 100 bits / $1) :"},
      style = "bold_label"
    }
    table.add {type = "label", name = shared.gui.amount_donation .. "_value", caption = "0"}
  end

  return table
end

gui.init_events = function()
  Event.register(
    -120,
    function()
      gui.update_tables()
    end
  )

  Event.register(
    defines.events.on_runtime_mod_setting_changed,
    function(event)
      if (event.setting:find("^" .. shared.prefix) ~= nil) then
        local player = game.players[event.player_index]
        gui.get_table(player, true)
      end
    end
  )
  Event.register(
    {defines.events.on_player_created, defines.events.on_player_joined_game},
    function(event)
      local player = Game.get_player(event)

      if player then
        gui.get_table(player, true)
        gui.update_player_table(player)
      end
    end
  )
end
gui.open_buy_dialog = function(player, chunk_count)
  chunk_count = chunk_count or 1

  local frame = player.gui.center
  local name_prefix = shared.prefix .. "_buy"

  if frame == nil or not frame.valid then
    return
  end

  local flow = frame[name_prefix .. "_flow"]
  if flow == nil or not flow.valid or reset then
    if flow then
      flow.destroy()
    end
    flow =
      frame.add {
      type = "flow",
      name = name_prefix .. "_flow",
      direction = "vertical"
    }
    flow.style.horizontally_stretchable = false
  end

  -- recreate
  local info_frame = flow[name_prefix .. "_frame"]
  if info_frame then
    info_frame.destroy()
  end
  info_frame =
    flow.add {
    type = "frame",
    name = name_prefix .. "_frame",
    direction = "vertical",
    style = mod_gui.frame_style,
    caption = "Buy chunk"
  }
  info_frame.style.horizontally_stretchable = true

  local chunks = #table.filter(global.control_data.donations.current_target, Core.VALID_FILTER)
  local setting = settings.global[shared.settings.chunk_cost].value
  local cost = chunks * setting

  local text =
    info_frame.add {
    type = "label",
    name = name_prefix .. "_msg",
    caption = {"gui.nilaus_s40_buy_credits_msg", chunks, cost}
  }
  text.style.top_padding = 2
  text.style.bottom_padding = 4

  local button_frame =
    info_frame.add {
    type = "flow",
    name = name_prefix .. "_button_frame",
    direction = "horizontal"
  }

  button_frame.add {
    type = "button",
    name = name_prefix .. "_cancel",
    caption = "Cancel",
    style = "back_button"
  }

  local confirm =
    button_frame.add {
    type = "button",
    name = name_prefix .. "_buy",
    caption = "Buy",
    style = "confirm_button"
  }

  if (cost > global.control_data.donations.amount) then
    text.caption = {
      "gui.nilaus_s40_buy_credits_msg_insufficient",
      chunks,
      cost,
      math.ceil(cost - global.control_data.donations.amount)
    }
    confirm.destroy()
  end
end

gui.on_buy = function(event)
  chunkManager.process_buy()
  gui.close_buy_dialog(event)
  gui.update_tables()
end

gui.on_buy_cancel = function(event)
  gui.close_buy_dialog(event)
end

gui.close_buy_dialog = function(event)
  local player = game.players[event.player_index]
  local name_prefix = shared.prefix .. "_buy"
  local frame = player.gui.center
  local flow = frame[name_prefix .. "_flow"]

  if (not flow or not flow.valid) then
    return
  end

  flow.destroy()
  global.control_data.donations.current_target = {}
  chunkManager.unhighlight_all()
end

gui.init_events = function()
  Event.register(
    -120,
    function()
      gui.update_tables()
    end
  )

  local name_prefix = shared.prefix .. "_buy"
  Gui.on_click(name_prefix .. "_cancel", gui.on_buy_cancel)
  Gui.on_click(name_prefix .. "_buy", gui.on_buy)

  Event.register(
    defines.events.on_runtime_mod_setting_changed,
    function(event)
      if (event.setting:find("^" .. shared.prefix) ~= nil) then
        local player = game.players[event.player_index]
        gui.get_table(player, true)
      end
    end
  )
end

return gui
